import os
import pathlib
import shutil
import subprocess
import sys
import sysconfig

import click
import dotenv

import PyQt5
import pyqt5_tools.badplugin
import pyqt5_tools.examplebuttonplugin
import pyqt5_tools.examples
import pyqt5_tools.examples.exampleqmlitem
import pyqt5_tools.tests.testbutton


fspath = getattr(os, 'fspath', str)


here = pathlib.Path(__file__).parent
bin = here/'Qt'/'bin'
example_path = str(
    pathlib.Path(pyqt5_tools.examplebuttonplugin.__file__).parent,
)
bad_path = str(
    pathlib.Path(pyqt5_tools.badplugin.__file__).parent,
)

pyqt5_root = pathlib.Path(PyQt5.__file__).parent

maybe_extension = {
    'linux': lambda name: name,
    'win32': lambda name: '{}.exe'.format(name),
    'darwin': lambda name: name,
}[sys.platform]


def pyqt5toolsinstalluic():
    destination = bin/'bin'
    destination.mkdir(parents=True, exist_ok=True)
    there = pathlib.Path(sys.executable).parent

    shutil.copy(
        src=str(there / maybe_extension('pyuic5')),
        dst=str(destination/maybe_extension('uic')),
    )


def load_dotenv():
    env_path = dotenv.find_dotenv(usecwd=True)
    if len(env_path) > 0:
        os.environ['DOT_ENV_DIRECTORY'] = str(pathlib.Path(env_path).parent)
        os.environ['SITE_PACKAGES'] = sysconfig.get_path('platlib')
        dotenv.load_dotenv(dotenv_path=env_path, override=True)


def create_env(reference):
    # TODO: uck, mutating
    load_dotenv()

    env = dict(reference)

    env.update(add_to_env_var_path_list(
        env=env,
        name='QT_PLUGIN_PATH',
        before=[],
        after=[fspath(here / 'Qt' / 'plugins')],
    ))
    # TODO: maybe also
    # PyQt5.QtCore.QLibraryInfo.location(
    #    PyQt5.QtCore.QLibraryInfo.PluginsPath,
    # )

    if sys.platform == 'linux':
        env.update(add_to_env_var_path_list(
            env=env,
            name='LD_LIBRARY_PATH',
            before=[''],
            after=[sysconfig.get_config_var('LIBDIR')],
        ))

    return env


def add_to_env_var_path_list(env, name, before, after):
    return {
        name: os.pathsep.join((
            *before,
            env.get(name, ''),
            *after
        ))
    }


def mutate_env_for_paths(env):
    env.update(add_to_env_var_path_list(
        env=env,
        name='PYTHONPATH',
        before=sys.path,
        after=[''],
    ))
    env.update(add_to_env_var_path_list(
        env=env,
        name='PATH',
        before=sys.path,
        after=[''],
    ))


def print_environment_variables(env, *variables):
    for name in variables:
        value = env.get(name)
        if value is None:
            print('{} is not set'.format(name))
        else:
            print('{}: {}'.format(name, value))


qt_debug_plugins_option = click.option(
    '--qt-debug-plugins/--no-qt-debug-plugins',
    help='Set QT_DEBUG_PLUGINS=1',
)


@click.command(
    context_settings={
        'ignore_unknown_options': True,
        'allow_extra_args': True,
    },
)
@click.pass_context
@click.option(
    '--widget-path',
    '-p',
    'widget_paths',
    help='Paths to be combined with PYQTDESIGNERPATH',
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    multiple=True,
)
@click.option(
    '--example-widget-path',
    help='Include the path for the pyqt5-tools example button ({})'.format(
        example_path,
    ),
    is_flag=True,
)
@click.option(
    '--designer-help',
    help='Pass through to get Designer\'s --help',
    is_flag=True,
)
@click.option(
    '--test-exception-dialog',
    help='Raise an exception to check the exception dialog functionality.',
    is_flag=True,
)
@qt_debug_plugins_option
def pyqt5designer(
        ctx,
        widget_paths,
        designer_help,
        example_widget_path,
        test_exception_dialog,
        qt_debug_plugins
):
    env = create_env(os.environ)

    extras = []
    widget_paths = list(widget_paths)

    if designer_help:
        extras.append('--help')

    if example_widget_path:
        widget_paths.append(example_path)

    if test_exception_dialog:
        widget_paths.append(bad_path)

    vars_to_print = [
        'PYQTDESIGNERPATH',
        'PYTHONPATH',
        'PATH',
        'QT_DEBUG_PLUGINS',
        'QT_PLUGIN_PATH',
        pyqt5_tools.tests.testbutton.test_path_env_var,
    ]

    if sys.platform == 'linux':
        vars_to_print.append('LD_LIBRARY_PATH')
        vars_to_print.append('DISPLAY')

    env.update(add_to_env_var_path_list(
        env=env,
        name='PYQTDESIGNERPATH',
        before=widget_paths,
        after=[''],
    ))

    mutate_env_for_paths(env)

    if qt_debug_plugins:
        env['QT_DEBUG_PLUGINS'] = '1'

    print_environment_variables(env, *vars_to_print)

    command = [
        str(bin / maybe_extension('designer')),
        *extras,
        *ctx.args,
    ]

    return subprocess.call(command, env=env)


qml2_import_path_option = click.option(
    '--qml2-import-path',
    '-p',
    'qml2_import_paths',
    help='Paths to be combined with QML2_IMPORT_PATH',
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    multiple=True,
)


def mutate_qml_path(env, paths):
    env.update(add_to_env_var_path_list(
        env=env,
        name='QML2_IMPORT_PATH',
        before=[*paths, fspath(pyqt5_root/'Qt'/'qml')],
        after=[''],
    ))


@click.command(
    context_settings={
        'ignore_unknown_options': True,
        'allow_extra_args': True,
    },
)
@click.pass_context
@qml2_import_path_option
@click.option(
    '--qmlscene-help',
    help='Pass through to get QML scene\'s --help',
    is_flag=True,
)
@qt_debug_plugins_option
@click.option(
    '--run-qml-example',
    help='Run the pyqt5-tools QML example',
    is_flag=True,
)
def pyqt5qmlscene(
        ctx,
        qml2_import_paths,
        qmlscene_help,
        qt_debug_plugins,
        run_qml_example,
):
    env = create_env(os.environ)
    extras = []

    if qmlscene_help:
        extras.append('--help')

    if run_qml_example:
        qml2_import_paths = qml2_import_paths + (fspath(here),)
        extras.append(fspath(
            pathlib.Path(pyqt5_tools.examples.__file__).parent/'qmlapp.qml'
        ))

    mutate_qml_path(env, paths=qml2_import_paths)
    mutate_env_for_paths(env)

    if qt_debug_plugins:
        env['QT_DEBUG_PLUGINS'] = '1'

    vars_to_print = [
        'QML2_IMPORT_PATH',
        'PYTHONPATH',
        'PATH',
        'QT_DEBUG_PLUGINS',
        'QT_PLUGIN_PATH',
        pyqt5_tools.examples.exampleqmlitem.test_path_env_var,
    ]

    if sys.platform == 'linux':
        vars_to_print.append('LD_LIBRARY_PATH')
        vars_to_print.append('DISPLAY')

    print_environment_variables(env, *vars_to_print)

    command = [
        str(bin / maybe_extension('qmlscene')),
        *extras,
        *ctx.args,
    ]

    return subprocess.call(command, env=env)


@click.command(
    context_settings={
        'ignore_unknown_options': True,
        'allow_extra_args': True,
    },
)
@click.pass_context
@qml2_import_path_option
@click.option(
    '--qmltestrunner-help',
    help='Pass through to get QML test runner\'s --help',
    is_flag=True,
)
@qt_debug_plugins_option
@click.option(
    '--test-qml-example',
    help='Test the pyqt5-tools QML example',
    is_flag=True,
)
def pyqt5qmltestrunner(
        ctx,
        qml2_import_paths,
        qmltestrunner_help,
        qt_debug_plugins,
        test_qml_example,
):
    env = create_env(os.environ)
    extras = []

    if qmltestrunner_help:
        extras.append('--help')

    if test_qml_example:
        qml2_import_paths = qml2_import_paths + (fspath(here),)
        extras.extend([
            '-input',
            fspath(
                pathlib.Path(pyqt5_tools.examples.__file__).parent/'qmltest.qml'
            ),
        ])

    mutate_qml_path(env, paths=qml2_import_paths)
    mutate_env_for_paths(env)

    if qt_debug_plugins:
        env['QT_DEBUG_PLUGINS'] = '1'

    vars_to_print = [
        'QML2_IMPORT_PATH',
        'PYTHONPATH',
        'PATH',
        'QT_DEBUG_PLUGINS',
        'QT_PLUGIN_PATH',
        pyqt5_tools.examples.exampleqmlitem.test_path_env_var,
    ]

    if sys.platform == 'linux':
        vars_to_print.append('LD_LIBRARY_PATH')
        vars_to_print.append('DISPLAY')

    print_environment_variables(env, *vars_to_print)

    command = [
        str(bin / maybe_extension('qmltestrunner')),
        *extras,
        *ctx.args,
    ]

    return subprocess.call(command, env=env)


# def designer():
#     env = create_env(os.environ)
#     return subprocess.call(
#         [
#             str(here/'Qt'/'bin'/'designer.exe'),
#             *sys.argv[1:],
#         ],
#         env=env,
#     )

# ----  start of generated wrapper entry points

def assistant():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/assistant.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def canbusutil():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/canbusutil.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def designer():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/designer.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def dumpcpp():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/dumpcpp.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def dumpdoc():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/dumpdoc.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def lconvert():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/lconvert.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def linguist():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/linguist.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def lprodump():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/lprodump.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def lrelease_pro():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/lrelease-pro.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def lrelease():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/lrelease.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def lupdate_pro():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/lupdate-pro.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def lupdate():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/lupdate.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def pixeltool():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/pixeltool.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qdbus():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qdbus.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qdbuscpp2xml():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qdbuscpp2xml.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qdbusviewer():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qdbusviewer.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qdbusxml2cpp():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qdbusxml2cpp.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qdistancefieldgenerator():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qdistancefieldgenerator.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qdoc():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qdoc.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qgltf():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qgltf.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qhelpgenerator():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qhelpgenerator.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qlalr():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qlalr.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qml():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qml.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qmlcachegen():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qmlcachegen.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qmleasing():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qmleasing.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qmlimportscanner():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qmlimportscanner.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qmllint():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qmllint.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qmlmin():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qmlmin.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qmlplugindump():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qmlplugindump.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qmlpreview():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qmlpreview.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qmlprofiler():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qmlprofiler.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qmlscene():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qmlscene.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qmltestrunner():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qmltestrunner.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qscxmlc():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qscxmlc.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qtattributionsscanner():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qtattributionsscanner.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qtdiag():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qtdiag.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qtpaths():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qtpaths.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qtplugininfo():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qtplugininfo.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def qvkgen():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/qvkgen.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def repc():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/repc.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def testcon():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/testcon.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def uic():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/uic.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def xmlpatterns():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/xmlpatterns.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )


def xmlpatternsvalidator():
    env = create_env(os.environ)
    return subprocess.call(
        [
            str(here/'Qt'/'bin/xmlpatternsvalidator.exe'),
            *sys.argv[1:],
        ],
        env=env,
    )



# ----  end of generated wrapper entry points

